#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Root of nti.intid.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__all__ = [
    'INTIIntIdAddedEvent',
    'INTIIntIdRemovedEvent',
    'NTIIntIdAddedEvent',
    'NTIIntIdRemovedEvent',

]

from zc.intid.interfaces import IAfterIdAddedEvent as INTIIntIdAddedEvent
from zc.intid.interfaces import IBeforeIdRemovedEvent as INTIIntIdRemovedEvent

from zc.intid.interfaces import AfterIdAddedEvent as NTIIntIdAddedEvent
from zc.intid.interfaces import BeforeIdRemovedEvent as NTIIntIdRemovedEvent



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains a :mod:`zc.intid.utility` derived utility for managing
intids. The primary reason to do this is to provide better exceptions,
and future proofing of behaviour.
"""
try:
    from Acquisition import aq_base
except ModuleNotFoundError: # pragma: no cover
    def aq_base(obj):
        return obj

import BTrees

from zc.intid.interfaces import RemovedEvent
from zc.intid.interfaces import IntIdInUseError

from zc.intid.utility import IntIds as _ZCIntIds

from zope import interface

from zope.event import notify as zope_notify

from zope.security.proxy import removeSecurityProxy as unwrap

from nti.intid.interfaces import IIntIds


import zope.deferredimport
zope.deferredimport.initialize()
zope.deferredimport.deprecated(
    "Import from zope.intid.interfaces instead",
    IntIdMissingError='zope.intid.interfaces:IntIdMissingError',
    ObjectMissingError='zope.intid.interfaces:ObjectMissingError')


logger = __import__('logging').getLogger(__name__)

@interface.implementer(IIntIds)
class IntIds(_ZCIntIds):
    """
    A integer ID utility that uses 64-bit values.

    Implements :class:`nti.intid.interfaces.IIntIds`
    """

    __name__ = None
    __parent__ = None

    #: The family of BTrees to use.
    family = BTrees.family64

    # Because this object stores IDs using attributes on the object,
    # it is important to be sure that the ID is not acquired
    # from a parent. Hence, all the methods use aq_base to unwrap
    # the object.
    #
    # Removing all proxies in general is more tricky; sometimes a
    # zope.container.contained.ContainedProxy is really what we want to register.
    # Fortunately, most proxies pass attributes on through to the underlying
    # object, in which case queryId will take either the proxy or the wrapped object;
    # alternatively, they define __slots__ and forbid new attributes

    def randomize(self):
        self._v_nextid = self._randrange(0, self.family.maxint)

    def queryId(self, ob, default=None):
        """
        Query for the id of the :func:`Acquisition.aq_base` of *ob*.

        .. note::

            If you pass a broken object (in the ZODB sense), this
            will hide that fact. We have to activate it, but if it is
            broken, we will not be able to. However, we catch
            :exc:`KeyError`, which is a superclass of the
            :exc:`~.POSKeyError` that gets thrown, so you cannot
            distinguish it at this point.

            We do not change this for backwards compatibility.
        """
        return _ZCIntIds.queryId(self, aq_base(ob), default)

    def register(self, ob, *unused_args, **unused_kwargs):
        """
        register(object) -> int

        Register the :func:`Acquisition.aq_base` of *object* and return the integer id.
        """
        result = _ZCIntIds.register(self, aq_base(ob))
        # 5 = ZODB.loglevels.TRACE
        logger.log(5, '%s was registered with intid %s', type(ob), result)
        return result

    def unregister(self, ob, *unused_args, **unused_kwargs):
        """
        unregister(object) -> None

        Unregister the :func:`Acquisition.aq_base` of *object*.
        """
        return _ZCIntIds.unregister(self, aq_base(ob))

    def getId(self, ob):
        """
        Get the id of the :func:`Acquisition.aq_base` of *ob*.

        See the note for :meth:`queryId`.
        """
        return _ZCIntIds.getId(self, aq_base(ob))
    get_id = getId

    def force_register(self, uid, ob, check=True):
        unwrapped = unwrap(aq_base(ob))
        if check and uid in self.refs:
            raise IntIdInUseError(ob)
        self.refs[uid] = unwrapped
        return uid
    forceRegister = force_register

    def force_unregister(self, uid, ob=None, notify=False, remove_attribute=True):
        if not uid in self.refs:
            raise KeyError(uid)
        if ob is not None:
            unwrapped = unwrap(aq_base(ob))
            if self.refs[uid] is not unwrapped:
                raise KeyError(ob)
        del self.refs[uid]
        if      remove_attribute \
            and ob is not None \
            and getattr(ob, self.attribute, None) is not None:
            setattr(ob, self.attribute, None)
        if notify and ob is not None:
            zope_notify(RemovedEvent(ob, self, uid))
    forceUnregister = force_unregister

    def __repr__(self):
        return "<%s.%s (%s) %s/%s>" % (self.__class__.__module__,
                                       self.__class__.__name__,
                                       self.attribute,
                                       self.__parent__, self.__name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=protected-access,too-many-public-methods

from hamcrest import is_
from hamcrest import none
from hamcrest import is_not
from hamcrest import assert_that

from persistent import Persistent

from zope import component
from zope import interface

from zope.intid.interfaces import IIntIds

from zope.location.interfaces import ILocation

from nti.intid.common import addIntId
from nti.intid.common import removeIntId

from nti.intid.tests import WithMockDS
from nti.intid.tests import IntIdTestCase

from nti.intid.tests import mock_db_trans


@interface.implementer(ILocation)
class User(Persistent):

    __parent__ = None

    def __init__(self, username):
        self.__name__ = username


class TestCommon(IntIdTestCase):

    @WithMockDS
    def test_add_remove(self):
        with mock_db_trans() as conn:
            intids = component.getUtility(IIntIds)
            user = User('ichigo@bleach.com')
            conn.add(user)
            addIntId(user)
            assert_that(intids.queryId(user), is_not(none()))

            removeIntId(user)
            assert_that(intids.queryId(user), is_(none()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=protected-access,too-many-public-methods
# pylint: disable=global-statement,broad-except
# pylint: disable=arguments-differ,no-member
import unittest

import BTrees

import zc.intid
import zope.intid
import zope.testing.cleanup
from zope.component.hooks import setHooks

from nti.site import testing as site_testing

class mock_db_trans(site_testing.persistent_site_trans):


    def on_connection_opened(self, conn):
        from nti.intid.utility import IntIds

        super().on_connection_opened(conn)

        folder = conn.root()[str(self.main_application_folder_name)]
        lsm = folder.getSiteManager()
        intids = IntIds('_ds_intid', family=BTrees.family64)
        intids.__name__ = '++etc++intids'
        intids.__parent__ = folder
        lsm.registerUtility(intids, provided=zope.intid.IIntIds)
        lsm.registerUtility(intids, provided=zc.intid.IIntIds)



WithMockDS = site_testing.uses_independent_db_site


class SharedConfiguringTestLayer(site_testing.SharedConfiguringTestLayer):

    set_up_packages = ('nti.intid',)

    @classmethod
    def setUp(cls):
        setHooks()
        cls.setUpPackages()

    @classmethod
    def tearDown(cls):
        cls.tearDownPackages()
        zope.testing.cleanup.cleanUp()

    @classmethod
    def testSetUp(cls, test=None):
        setHooks()

    @classmethod
    def testTearDown(cls):
        pass


class IntIdTestCase(unittest.TestCase):
    layer = SharedConfiguringTestLayer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Common utilities.

Currently, there is nothing here.

See :mod:`zc.intid.subscribers`
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from zc.intid.subscribers import addIntIdSubscriber
from zc.intid.subscribers import removeIntIdSubscriber


def addIntId(ob, event=None):
    """
    Deprecated alias for :func:`zc.intid.subscribers.addIntIdSubscriber`
    """
    addIntIdSubscriber(ob, event)
add_intid = addIntId


def removeIntId(ob, event=None):
    """
    Deprecated alias for :func:`zc.intid.subscribers.removeIntIdSubscriber`
    """
    removeIntIdSubscriber(ob, event)
remove_intid = removeIntId



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=protected-access,too-many-public-methods
# pylint: disable=attribute-defined-outside-init

from hamcrest import is_
from hamcrest import none
from hamcrest import is_not
from hamcrest import raises
from hamcrest import calling
from hamcrest import has_length
from hamcrest import assert_that
from hamcrest import has_property
does_not = is_not

from nti.testing.matchers import validly_provides
from nti.testing.matchers import verifiably_provides

from nti.testing.base import AbstractTestBase

import struct

from zc.intid import IIdAddedEvent
from zc.intid import IIdRemovedEvent

from zc.intid.interfaces import IntIdInUseError

from zope import interface

from zope.component import eventtesting

from zope.location.interfaces import ILocation

from persistent import Persistent

from nti.intid.interfaces import IIntIds

from nti.intid.utility import IntIds


@interface.implementer(ILocation)
class P(Persistent):
    pass


class ConnectionStub(object):
    next = 1

    database_name = 'ConnectionStub'

    def db(self):
        return self

    def add(self, ob):
        ob._p_jar = self
        ob._p_oid = struct.pack(">I", self.next)
        self.next += 1

    def register(self, *args, **kwargs):
        pass


class TestUtility(AbstractTestBase):

    def setUp(self):
        super().setUp()
        eventtesting.setUp()

    def test_interface(self):
        u = IntIds("_ds_id")
        assert_that(u, validly_provides(IIntIds))
        assert_that(u, verifiably_provides(IIntIds))
        assert_that(repr(u), is_not(none()))

    def test_non_keyreferences(self):
        u = IntIds("_ds_id")
        obj = object()
        assert_that(u.queryId(obj), is_(none()))
        assert_that(u.unregister(obj), is_(none()))
        assert_that(calling(u.getId).with_args(obj), raises(KeyError))

    def test_event(self):
        u = IntIds("_ds_id")

        obj = P()
        stub = ConnectionStub()
        stub.add(obj)
        assert_that(stub.db(), is_(stub))

        assert_that(obj,
                    has_property('_p_jar', is_(stub)))

        count = 1
        assert_that(calling(u.getId).with_args(obj), raises(KeyError))
        assert_that(calling(u.getId).with_args(P()), raises(KeyError))

        assert_that(u.queryId(obj), is_(none()))
        assert_that(u.queryId(obj, 42), is_(42))
        assert_that(u.queryId(P(), 42), is_(42))
        assert_that(u.queryObject(42), is_(none()))
        assert_that(u.queryObject(42, obj), is_(obj))

        uid = u.register(obj)
        assert_that(u.getObject(uid), is_(obj))
        assert_that(u.queryObject(uid), is_(obj))
        assert_that(u.getId(obj), is_(uid))
        assert_that(u.queryId(obj), is_(uid))
        assert_that(eventtesting.getEvents(IIdAddedEvent), has_length(count))

        uid2 = u.register(obj)
        assert_that(uid, is_(uid2))

        u.unregister(obj)
        assert_that(calling(u.getObject).with_args(uid), raises(KeyError))
        assert_that(calling(u.getId).with_args(obj), raises(KeyError))
        assert_that(eventtesting.getEvents(IIdRemovedEvent), has_length(count))

        next_id = u._v_nextid
        assert_that(u.randomize(), is_not(next_id))

    def test_force(self):
        u = IntIds("_ds_id")
        obj = P()
        obj._ds_id = 100
        stub = ConnectionStub()
        stub.add(obj)

        u.force_register(100, obj)
        assert_that(calling(u.force_register).with_args(100, P()),
                    raises(IntIdInUseError))

        assert_that(calling(u.force_unregister).with_args(200, obj),
                    raises(KeyError))

        assert_that(calling(u.force_unregister).with_args(100, P()),
                    raises(KeyError))

        u.force_unregister(100, obj, True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=protected-access,too-many-public-methods
import pickle

from hamcrest import is_
from hamcrest import none
from hamcrest import is_not
from hamcrest import not_none
from hamcrest import assert_that
from hamcrest import has_property
from hamcrest import has_length

import BTrees.OOBTree # pylint:disable=no-name-in-module,import-error

from persistent import Persistent
from zope import interface
from zope.location.interfaces import ILocation

from nti.wref.interfaces import ICachingWeakRef
from nti.wref.interfaces import IWeakRef
from nti.wref.interfaces import IWeakRefToMissing


from nti.testing.matchers import validly_provides as verifiably_provides

from nti.intid import wref
from nti.intid.tests import IntIdTestCase
from nti.intid.tests import WithMockDS
from nti.intid.tests import mock_db_trans

root_name = str(mock_db_trans.main_application_folder_name)

@interface.implementer(ILocation)
class User(Persistent):

    __parent__ = None

    def __init__(self, username):
        self.__name__ = username


class TestIntidWref(IntIdTestCase):

    def _create_user(self, username, conn=None):
        user = User(username)
        if conn is not None:
            folder = conn.root()[root_name]
            folder[username] = user
        return user

    @WithMockDS
    def test_pickle(self):
        with mock_db_trans() as conn:
            user = self._create_user('sjohnson@nextthought.com', conn)
            ref = wref.WeakRef(user)

            assert_that(ref, has_property('_v_entity_cache', user))

            copy = pickle.loads(pickle.dumps(ref))

            assert_that(copy, has_property('_v_entity_cache', none()))

            assert_that(copy(), is_(user))
            assert_that(ref, is_(copy))
            assert_that(copy, is_(ref))
            assert_that(repr(copy), is_(repr(ref)))
            assert_that(hash(copy), is_(hash(ref)))

            assert_that(ref, verifiably_provides(IWeakRef))
            assert_that(ref, verifiably_provides(ICachingWeakRef))
            assert_that(ref, verifiably_provides(IWeakRefToMissing))

    @WithMockDS
    def test_missing(self):
        with mock_db_trans() as conn:
            user = self._create_user('sjohnson@nextthought.com', conn)
            # Cannot find with invalid intid
            ref = wref.WeakRef(user)
            setattr(ref, '_v_entity_cache', None)
            setattr(ref, '_entity_id', -1)
            assert_that(ref(), is_(none()))
            assert_that(ref.make_missing_ntiid(),
                        is_('tag:nextthought.com,2011-10:Missing'))
            # cannot find with invalid oid
            ref = wref.WeakRef(user)
            assert_that(ref, has_property('_entity_oid', not_none()))
            setattr(ref, '_v_entity_cache', None)
            setattr(ref, '_entity_oid', -1)
            assert_that(ref(), is_(none()))

            # Find it with oid of None (but valid intid)
            ref = wref.WeakRef(user)
            assert_that(ref, has_property('_entity_oid', not_none()))
            setattr(ref, '_v_entity_cache', None)
            setattr(ref, '_entity_oid', None)
            assert_that(ref(), is_(user))

            # Caching can be controlled
            ref = wref.WeakRef(user)
            setattr(ref, '_entity_id', -1)
            setattr(ref, '_entity_oid', None)
            assert_that(ref(), is_(user))  # From cache

            assert_that(ref(allow_cached=False), is_(none()))  # not from cache

    @WithMockDS
    def test_no_caching(self):
        with mock_db_trans() as conn:
            user = self._create_user('sjohnson@nextthought.com', conn)
            ref = wref.NoCachingArbitraryOrderableWeakRef(user)
            assert_that(ref(), is_not(none()))

    @WithMockDS
    def test_in_btree(self):
        import warnings
        with mock_db_trans() as conn:
            user_1 = self._create_user('sjohnson@nextthought.com', conn)
            user_2 = self._create_user('sjohnson2@nextthought.com', conn)

            for clazz in (wref.NoCachingArbitraryOrderableWeakRef,
                          wref.ArbitraryOrderableWeakRef,
                          wref.WeakRef):
                bt = BTrees.OOBTree.OOBTree() # pylint:disable=no-member

                ref_1 = clazz(user_1)
                ref_2 = clazz(user_2)
                with warnings.catch_warnings(record=True):
                    bt[ref_1] = 1
                    bt[ref_2] = 2

                    assert_that(bt[ref_1], is_(1))
                    assert_that(bt[ref_2], is_(2))

    @WithMockDS
    def test_eq_ne(self):
        with mock_db_trans() as conn:
            user_1 = self._create_user('sjohnson@nextthought.com', conn)
            user_2 = self._create_user('sjohnson2@nextthought.com', conn)

            ref_1 = wref.ArbitraryOrderableWeakRef(user_1)
            ref_2 = wref.ArbitraryOrderableWeakRef(user_2)

            assert_that(ref_1, is_(ref_1))
            assert_that(ref_2, is_not(ref_1))
            assert_that(ref_1, is_not(ref_2))

            assert_that(ref_1 != ref_1, is_(False)) # pylint:disable=comparison-with-itself
            assert_that(ref_1 != ref_2, is_(True))
            assert_that(ref_1 != object(), is_(True))

    @WithMockDS
    def test_non_orderable_weakref_generates_warnings(self):
        import warnings
        import operator
        with mock_db_trans() as conn:
            user_1 = self._create_user('sjohnson@nextthought.com', conn)
            user_2 = self._create_user('sjohnson2@nextthought.com', conn)

            ref_1 = wref.WeakRef(user_1)
            ref_2 = wref.WeakRef(user_2)

            with warnings.catch_warnings(record=True) as warns:
                operator.lt(ref_1, ref_2)

            assert_that(warns, has_length(1))

            with warnings.catch_warnings(record=True) as warns:
                operator.gt(ref_1, ref_2)

            assert_that(warns, has_length(1))

    @WithMockDS
    def test_arbitrary_orderable_weakref(self):
        import operator
        with mock_db_trans() as conn:
            user_1 = self._create_user('sjohnson@nextthought.com', conn)
            user_2 = self._create_user('sjohnson2@nextthought.com', conn)

            ref_1 = wref.ArbitraryOrderableWeakRef(user_1)
            ref_2 = wref.ArbitraryOrderableWeakRef(user_2)

            operator.lt(ref_1, ref_2)
            operator.gt(ref_1, ref_2)

    @WithMockDS
    def test_no_dict(self):
        with mock_db_trans() as conn:
            user_1 = self._create_user('sjohnson@nextthought.com', conn)

            for clazz in (wref.NoCachingArbitraryOrderableWeakRef,
                          wref.ArbitraryOrderableWeakRef,
                          wref.WeakRef):
                __traceback_info__ = clazz
                ref = clazz(user_1)
                self.assertFalse(hasattr(ref, '__dict__'))
                with self.assertRaises(AttributeError):
                    setattr(ref, 'arbitrary_attribute', None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interfaces for working with integer IDs.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from zc.intid.interfaces import IIntIdsSubclass
from zc.intid.interfaces import IIntIds as IZCIIntIds

from zope.location.interfaces import IContained

import zope.deferredimport
zope.deferredimport.initialize()
zope.deferredimport.deprecated(
    "Import from zope.intid.interfaces instead",
    IntIdMissingError='zope.intid.interfaces:IntIdMissingError',
    ObjectMissingError='zope.intid.interfaces:ObjectMissingError')

zope.deferredimport.deprecated(
    "Import from zc.intid.interfaces instead",
    IntIdAlreadyInUseError='zc.intid.interfaces:IntIdInUseError',
    IIntIdEvent='zc.intid.interfaces:ISubscriberEvent',
    IIntIdAddedEvent='zc.intid.interfaces:IAfterIdAddedEvent',
    IIntIdRemovedEvent='zc.intid.interfaces:IBeforeIdRemovedEvent',
    IntIdAddedEvent='zc.intid.interfaces:AfterIdAddedEvent',
    IntIdRemovedEvent='zc.intid.interfaces:BeforeIdRemovedEvent')

# pylint:disable=no-method-argument,no-self-argument
# pylint:disable=too-many-ancestors

class IIntIds(IZCIIntIds, IIntIdsSubclass, IContained):
    """
    Advanced extensions to an integer ID catalog.

    Normal users will have no need of this interface.
    """

    def randomize():
        """
        Randomize the next id.

        Do not use this method.
        """

    def force_register(uid, ob, check=True):
        """
        Register an object.

        Do not use this method.

        :param uid: Registration id
        :param ob: Object to register
        :param check: Validation check flag
        """

    def force_unregister(uid, ob=None, notify=False, remove_attribute=True):
        """
        Unregister an object.

        Do not use this method.

        :param uid: Id to unregister
        :param ob: Obj to unregister [optional]
        :param notify: Flag to trigger an ``IIdRemovedEvent``
        :param notremove_attribute: Flag to remove intid attribute
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=protected-access,too-many-public-methods

import sys
import types
import unittest

from zope.dottedname import resolve as dottedname


class TestModules(unittest.TestCase):

    def test_import_interfaces(self):
        dottedname.resolve('nti.intid.interfaces')

    def test_import_containers(self):
        location = 'nti.containers'
        if location not in sys.modules:
            for name in ('', '.datastructures'):
                new_loc = location + name
                sys.modules[new_loc] = types.ModuleType(new_loc, "Created module")
        dottedname.resolve('nti.intid.containers')



import codecs
from setuptools import setup
from setuptools import find_namespace_packages

entry_points = {
    'console_scripts': [
    ],
}

TESTS_REQUIRE = [
    'coverage',
    'ZODB',
    'nti.site',
    'nti.testing',
    'persistent',
    'transaction',
    'zope.dottedname',
    'zope.site',
    'zope.testrunner',
]


def _read(fname):
    with codecs.open(fname, encoding='utf-8') as f:
        return f.read()


setup(
    name='nti.intid',
    version='1.1.0',
    author='Jason Madden',
    author_email='jason@nextthought.com',
    description="Extensions for Zope3's intids.",
    long_description=(
        _read('README.rst')
        + '\n\n'
        + _read("CHANGES.rst")
    ),
    license='Apache',
    keywords='intid zope3',
    classifiers=[
        'Framework :: Zope3',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    url="https://github.com/OpenNTI/nti.intid",
    zip_safe=True,
    packages=find_namespace_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'BTrees',
        'nti.externalization',
        'nti.ntiids',
        'nti.wref',
        'zc.intid',
        'zope.component',
        'zope.deferredimport',
        'zope.deprecation',
        'zope.event',
        'zope.interface',
        'zope.intid',
        'zope.keyreference',
        'zope.location',
        'zope.security',
    ],
    extras_require={
        'test': TESTS_REQUIRE,
        'docs': [
            'Sphinx',
            'repoze.sphinx.autointerface',
            'sphinx_rtd_theme',
        ],
    },
    entry_points=entry_points,
    python_requires=">=3.12",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Weak references to things with intids. These serve the same purpose as general
persistent weak references from :mod:`persistent.wref`, but are specific
to things with intids, and do not keep the object alive or accessible once
the object is removed from the intid catalog
(whereas weak refs do until such time as the database is GC'd).

.. $Id$
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import warnings
import functools

from zc.intid import IIntIds

from zope import component
from zope import interface

from nti.externalization.integer_strings import to_external_string

from nti.ntiids.ntiids import TYPE_MISSING
from nti.ntiids.ntiids import make_ntiid

from nti.wref.interfaces import ICachingWeakRef
from nti.wref.interfaces import IWeakRefToMissing

logger = __import__('logging').getLogger(__name__)


class _AbstractWeakRef(object):
    """
    A weak reference to a content object (generally, anything
    with an intid). Call this object to return either the
    object, or if the object has gone away, ``None``.

    Subclasses need to support three properties: ``_entity_id``,
    ``_entity_oid``, and ``_v_entity_cache``. The first is the intid.
    Because intids can be reused, we also include a backup, the
    object's OID (if it has one) in the second. This isn't foolproof,
    but should to pretty good. The third is used for caching.
    Subclasses will need to decide how to implement ``__getstate__``
    and ``__setstate__``. If subclasses need a ``__dict__``, they will have
    to declare that in ``__slots__.``
    """

    __slots__ = ('_entity_id', '_entity_oid', '_v_entity_cache')

    def __init__(self, content_object):
        self._entity_id = component.getUtility(IIntIds).getId(content_object)
        # _v_entity_cache is a volatile attribute. It's either None, meaning we have
        # no idea, the resolved object, or False
        self._v_entity_cache = content_object
        self._entity_oid = getattr(content_object, '_p_oid', None)

    def _cached(self, allow_cached):
        if allow_cached and self._v_entity_cache is not None:
            # Notice that we must explicitly check for _v_entity_cache being
            # not False, because we use False to represent a failed lookup. (But
            # we cannot use just a simple bool comparison because that fails for
            # something like an empty container)
            return self._v_entity_cache if self._v_entity_cache is not False else None

        try:
            result = component.getUtility(IIntIds).getObject(self._entity_id)
        except KeyError:
            result = None

        if self._entity_oid is not None:
            result_oid = getattr(result, '_p_oid', None)
            if result_oid is None or result_oid != self._entity_oid:
                result = None

        if allow_cached:  # only perturb the state if we are allowed to
            if result is not None:
                self._v_entity_cache = result
            else:
                self._v_entity_cache = False

        return result

    def __call__(self, allow_cached=True):
        return self._cached(allow_cached)

    def __eq__(self, other):
        if self is other:
            return True
        # pylint: disable=protected-access
        try:
            if self._entity_id == other._entity_id:
                return self._entity_oid == other._entity_oid \
                    or self._entity_oid is None
            return False
        except AttributeError:  # pragma: no cover
            return NotImplemented

    def __ne__(self, other):
        if self is other:
            return False
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return eq
        return not eq

    def __hash__(self):
        return hash((self._entity_id, self._entity_oid))

    def __repr__(self):
        return "<%s.%s %s>" % (self.__class__.__module__,
                               self.__class__.__name__,
                               self._entity_id)

    def make_missing_ntiid(self):
        eid = self._entity_id
        # This intid is probably no longer used, but we have no guarantee
        # of that. We do some trivial manipulation on it to make it less
        # obvious what it is, and less likely to come into the system when
        # we don't want it to
        eid = to_external_string(eid)
        # base64 might be nice, but that doesn't play well with ntiids
        return make_ntiid(nttype=TYPE_MISSING, specific=eid)


@interface.implementer(IWeakRefToMissing, ICachingWeakRef)
class WeakRef(_AbstractWeakRef):
    """
    A weak reference to a content object (generally, anything
    with an intid). Call this object to return either the
    object, or if the object has gone away, ``None``.

    Note that this is not a persistent object. It is not mutable (and
    it's also very tiny), and it has value semantics, so there's very
    little need to be persistent. This means it's suitable for use in
    OOSet objects and OOBTree objects as keys (if subclassed to provide
    total ordering!)
    """

    __slots__ = ()

    def __getstate__(self):
        return self._entity_id, self._entity_oid

    def __setstate__(self, state):
        self._entity_id, self._entity_oid = state
        self._v_entity_cache = None

    # We are not orderable. Hopefully there is no data like this in the real
    # world, but we did have tests relying on it. Try to catch it here,
    # doing what Python 2 would have done itself (Python 3 would raise errors).

    _unorderable_message = (
        "WeakRef is not totally orderable. Storing this in "
        "BTrees is a bad idea. Seeing this warning come from anything other than a "
        "test module is an error."
    )

    def __lt__(self, other):
        warnings.warn(self._unorderable_message, stacklevel=2)
        return id(self) < id(other)

    def __gt__(self, other):
        warnings.warn(self._unorderable_message, stacklevel=2)
        return id(self) > id(other)


@functools.total_ordering
class ArbitraryOrderableWeakRef(WeakRef):
    """
    A subclass of :class:`WeakRef` that is orderable in a completely arbitrary
    way (based simply on intids).
    """

    __slots__ = ()

    def __lt__(self, other):
        # pylint: disable=protected-access
        try:
            return (self._entity_id, self._entity_oid) < (other._entity_id, other._entity_oid)
        except AttributeError:  # pragma: no cover
            return NotImplemented

    def __gt__(self, other):
        # pylint: disable=protected-access
        try:
            return (self._entity_id, self._entity_oid) > (other._entity_id, other._entity_oid)
        except AttributeError:  # pragma: no cover
            return NotImplemented


class NoCachingArbitraryOrderableWeakRef(ArbitraryOrderableWeakRef):
    """
    Does not allow caching.
    """
    __slots__ = ()
    def __call__(self, unused_allow_cached=False):
        return ArbitraryOrderableWeakRef.__call__(self, allow_cached=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Subscribers for events.

These are configured by loading this packages's ``configure.zcml``.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from zope import component

from zope.component import handle

from zc.intid.interfaces import ISubscriberEvent


@component.adapter(ISubscriberEvent)
def subscriberEventNotify(event):
    """
    Event subscriber to dispatch
    :class:`zc.intid.interfaces.ISubscriberEvent` to interested
    adapters.

    The adapters are registered on ``event.object, event``.
    """
    handle(event.object, event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.. $Id$
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import zope.deprecation
zope.deprecation.moved('nti.containers.datastructures')
